"""PySpark-shaped ``Window`` / ``WindowSpec`` (see :mod:`pydantable.window_spec`)."""

from __future__ import annotations

from pydantable.window_spec import Window, WindowSpec

__all__ = ["Window", "WindowSpec"]
